"""
Amatino API Python Bindings
Decodable Module
Author: hugh@amatino.io
"""
from amatino.entity import Entity
from json import loads
from typing import TypeVar
from typing import Type
from typing import List
from typing import Any
from amatino.unexpected_response_type import UnexpectedResponseType

T = TypeVar('T', bound='Decodable')


class Decodable:
    """
    Abstract class defining an interface for objects that may be decoded from
    serialised data, and are associated with an Entity
    """

    @classmethod
    def decode(cls: Type[T], entity: Entity, data: Any) -> T:
        raise NotImplementedError

    @classmethod
    def decode_many(cls: Type[T], entity: Entity, data: Any) -> List[T]:
        if not isinstance(data, list):
            raise UnexpectedResponseType(data, list)
        return [cls.decode(entity, o) for o in data]

    @classmethod
    def deserialise(cls: Type[T], entity: Entity, data: str) -> T:
        return cls.decode(entity, loads(data))

    @classmethod
    def deserialise_many(cls: Type[T], entity: Entity, data: str) -> List[T]:
        return cls.decode_many(entity, loads(data))



"""
Amatino API Python Bindings
Immutable Module
Author: hugh@amatino.io
"""
from amatino.amatino_error import AmatinoError


class Immutable(property):
    """
    An extension of the inbuit `property` class, which enforces immutability
    upon object properties.
    """
    _MESSAGE = """
    Amatino object instances may not be mutated via properties. Use the
    .update() method to make changes.
    """

    def __init__(self, fget) -> None:

        super().__init__(
            fget,
            self._set_error,
            self._del_error,
            None
        )

    def _set_error(self, _1, _2) -> None:
        raise AmatinoError(self._MESSAGE)

    def _del_error(self, _) -> None:
        raise AmatinoError(self._MESSAGE)



"""
Amatino API Python Bindings
Recursive Balance Test Module
Author: hugh@amatino.io
"""
from amatino.tests.primary.transaction import TransactionTest
from amatino import RecursiveBalance
from decimal import Decimal
from amatino import Account
from amatino import Transaction
from amatino import Entry
from amatino import Side
from datetime import datetime

NAME = 'Retrieve a Recursive Balance'


class RecursiveBalanceTest(TransactionTest):
    """Test the Recursive Balance object"""

    def __init__(self, name=NAME) -> None:

        super().__init__(name)
        return

    def execute(self) -> None:

        try:

            asset_child = Account.create(
                self.session,
                self.entity,
                'Test Asset Child',
                self.asset.am_type,
                self.asset.denomination,
                parent=self.asset
            )

            self.create_transaction(amount=Decimal(42))
            self.create_transaction(amount=Decimal(18))

            Transaction.create(
                self.session,
                self.entity,
                datetime.utcnow(),
                [
                    Entry(Side.debit, Decimal(5), asset_child),
                    Entry(Side.credit, Decimal(5), self.liability)
                ],
                asset_child.denomination
            )

        except Exception as error:
            self.record_failure(error)
            return

        try:
            balance = RecursiveBalance.retrieve(self.entity, self.asset)
        except Exception as error:
            self.record_failure(error)
            return

        if balance.magnitude != Decimal(65):
            message = 'Unexpected magnitude: ' + str(balance.magnitude)
            self.record_failure(message)
            return

        self.record_success()
        return



"""
Amatino API Python Bindings
Account Test Module
Author: hugh@amatino.io
"""
from datetime import datetime
from amatino import Transaction
from amatino import Side
from amatino import Entry
from decimal import Decimal
from amatino.tests.primary.account import AccountTest
from amatino import AMType
from urllib.error import HTTPError
from typing import Optional

NAME = 'Create, retrieve, update, delete a Transaction'


class TransactionTest(AccountTest):
    """Test the Transaction object"""

    def __init__(self, name=NAME) -> None:

        super().__init__(name)
        self.asset = self.create_account(AMType.asset, 'Test Asset')
        self.liability = self.create_account(AMType.liability, 'Test Liability')
        return

    def create_transaction(
        self,
        time: Optional[datetime] = None,
        amount: Optional[Decimal] = None
    ) -> Transaction:

        if time is None:
            time = datetime.utcnow()
        if amount is None:
            amount = Decimal(10)

        transaction = Transaction.create(
            self.session,
            self.entity,
            datetime.utcnow(),
            [
                Entry(Side.debit, amount, self.asset),
                Entry(Side.credit, amount, self.liability)
            ],
            self.usd,
            'Test transaction'
        )

        return transaction

    def execute(self) -> None:

        try:
            transaction = self.create_transaction()
        except Exception as error:
            self.record_failure(error)
            return

        assert isinstance(transaction, Transaction)

        try:
            transaction = Transaction.retrieve(
                self.session,
                self.entity,
                transaction.id_,
                self.asset.denomination
            )
        except Exception as error:
            self.record_failure(error)
            return

        updated_description = 'Updated test description goodness'

        try:
            transaction = transaction.update(description=updated_description)
        except Exception as error:
            self.record_failure(error)
            return

        if not transaction.description == updated_description:
            self.record_failure('Transaction description not updated')
            return

        try:
            transaction.delete()
        except Exception as error:
            self.record_failure(error)
            return

        try:
            Transaction.retrieve(
                self.session,
                self.entity,
                transaction.id_,
                self.asset.denomination
            )
        except Exception as error:
            if isinstance(error, HTTPError) and error.code == 404:
                self.record_success()
                return
            self.record_failure(error)

        return



"""
Amatino API Python Bindings
Global Unit Test Module
Author: hugh@amatino.io
"""
from amatino.tests.ancillary.session import SessionTest
from amatino import GlobalUnit

USD_UNIT_ID = 5


class GlobalUnitTest(SessionTest):
    """Test the Global Unit object"""

    def __init__(self, name='Retrieve a Global Unit') -> None:

        super().__init__(name)
        self.create_session()

        return

    def execute(self) -> None:

        try:
            usd = GlobalUnit.retrieve(self.session, USD_UNIT_ID)
        except Exception as error:
            self.record_failure(error)
            return

        if not isinstance(usd.code, str):
            self.record_failure('Unexpected code type ' + str(type(usd.code)))
            return

        if not isinstance(usd.name, str):
            self.record_failure('Unexpected name type ' + str(type(usd.name)))
            return

        self.record_success()
        return



"""
Amatino API Python Bindings
AM Type Module
Author: hugh@amatino.io
"""
from enum import Enum


class AMType(Enum):
    """
    Double-entry accounting divides accounts into five fundamental types:
    Assets, liabilities, equities, incomes and expenses. Throughout Amatino,
    these five constants are referred to as Types.

    Because 'Type' is such a loaded word in the programming world, Amatino
    Python name-spaces Types as AMType.

    You will most often encounter Types when creating Accounts. Erstwhile, most
    of the work that Types do occurs behind the scenes, and you won't need to
    interact with them directly.

    Under the hood, Types allow Amatino to maintain the Fundamental Double-Entry
    Equality.
    """
    asset = 1
    liability = 2
    equity = 3
    income = 4
    expense = 5



"""
Amatino API Python Bindings
Recursive Ledger Module
Author: hugh@amatino.io
"""
from amatino.ledger import Ledger


class RecursiveLedger(Ledger):
    """
    A Recursive Ledger is a list of Transactions from the perspective of a
    particular Account, and all the Transactions party to all child Accounts
    of that Account. Recursive Ledgers are ordered by Transaction time, and
    include a running Account Balance for every line.

    If you request a Recursive Ledger in a unit other than the target Account
    native unit, or if child Accounts feature heterogenous units, Amatino will
    compute and return unrealised gains and losses. Find out more in the
    Unrealised Gains and Losses article

    Amatino will return a maximum total of 1,000 Ledger Rows per retrieval
    request. If the Ledger you define spans more than 1,000 rows, it will be
    broken into pages you can retrieve seperately.
    """
    _PATH = '/accounts/ledger/recursive'



"""
Amatino API Python Bindings
Denominated Module
Author: hugh@amatino.io
"""
from amatino.denomination import Denomination
from amatino.custom_unit import CustomUnit
from amatino.global_unit import GlobalUnit
from amatino.internal.immutable import Immutable
from amatino.entity import Entity
from typing import TypeVar

T = TypeVar('T', bound='Denominated')


class Denominated:
    """
    Abstract class defining an interface for objects that are
    denominated in an Amatino unit of account, i.e. in a Custom Unit or
    Global Unit. Provides default functionality for determining what
    unit denominates an object. Provides default caching capacity
    for retrieved Custom Units and Global Units such that repeated requests
    for them from properties will not result in extra synchronous calls
    to the Amatino API
    """

    global_unit_id = NotImplemented
    custom_unit_id = NotImplemented
    denomination = Immutable(lambda s: s._denomination())
    entity = NotImplemented

    _denominated_cached_custom_unit = None
    _denominated_cached_global_unit = None

    def _denomination(self) -> Denomination:
        """Return the unit denominating this object"""
        if self.global_unit_id == NotImplemented:
            raise NotImplementedError('Implement .global_unit_id property')
        if self.custom_unit_id == NotImplemented:
            raise NotImplementedError('Implement .custom_unit_id property')
        if not isinstance(self.entity, Entity):
            raise NotImplementedError('Implement .entity property')

        if self.global_unit_id is not None and self.custom_unit_id is not None:
            raise AssertionError('Both global & custom units supplied!')

        if self.global_unit_id is not None:
            assert isinstance(self.global_unit_id, int)
        if self.custom_unit_id is not None:
            assert isinstance(self.custom_unit_id, int)

        if self.global_unit_id is not None:
            assert isinstance(self.global_unit_id, int)
            if self._denominated_cached_global_unit is not None:
                return self._denominated_cached_global_unit
            global_unit = GlobalUnit.retrieve(
                self.entity.session,
                self.global_unit_id
            )
            self._denominated_cached_global_unit = global_unit
            return global_unit

        if self._denominated_cached_custom_unit is not None:
            return self._denominated_cached_custom_unit
        custom_unit = CustomUnit.retrieve(
            self.entity,
            self.entity.session,
            self.custom_unit_id
        )
        self._denominated_cached_custom_unit = custom_unit
        return custom_unit
